#!/usr/bin/python
# -*- coding: UTF-8 -*-

# Copyright (c) 2012 Fabien LOISON <http://www.flogisoft.com>
#
# This program is free software. It comes without any warranty, to
# the extent permitted by applicable law. You can redistribute it
# and/or modify it under the terms of the Do What The Fuck You Want
# To Public License, Version 2, as published by Sam Hocevar. See
# http://sam.zoy.org/wtfpl/COPYING for more details.


FILES = [
    ["user-manual.md", u"2048 - Guide de l'utilisateur"],
    ["technical-manual.md", u"2048 - Manuel technique"],
    ]


import os
import sys
import codecs
from StringIO import StringIO

import markdown
import jinja2
import weasyprint


def build_doc(inputfile, title, outputhtml, outputpdf):
    sys.stdout.write("Processing %s...\n" % inputfile)

    #### HTML ####
    sys.stdout.write("  * HTML output... ")
    # Markdown To HTML
    output = StringIO()
    markdown.markdownFromFile(input=inputfile, output=output, encoding="utf-8")
    # Insert in template
    env = jinja2.Environment(loader=jinja2.FileSystemLoader("templates"))
    template = env.get_template("layout.html")
    with open(outputhtml, "w") as out:
        content = output.getvalue().decode("utf-8")
        content = content.replace("<p>TODO</p>", '<p><span style="background: #DC9700; border-radius: 5px; padding: 2px 5px; font-weight: bold; color: #fff;">TODO</span></p>') #FIXME
        content = content.replace("<p>FIXME</p>", '<p><span style="background: #DC9700; border-radius: 5px; padding: 2px 5px; font-weight: bold; color: #fff;">FIXME</span></p>') #FIXME
        out.write(template.render(title=title, content=content).encode("utf-8"))
    sys.stdout.write("Done!\n")

    #### PDF ####
    sys.stdout.write("  * PDF output...  ")
    weasyprint.HTML(outputhtml).write_pdf(outputpdf)
    sys.stdout.write("Done!\n")


for file_, title in FILES:
    build_doc(
            os.path.join("src", file_),
            title,
            os.path.join("html", file_.replace(".md", ".html")),
            os.path.join("pdf", file_.replace(".md", ".pdf")))
