var e_board = document.getElementById("board");
var e_overlayGameOver = document.getElementById("overlay-gameover");
var e_overlayGoodGame = document.getElementById("overlay-goodgame");


var $Class = Class.$extend({
    __init__: function() {
        for (var prop in this) {
            if (prop.indexOf("get") == 0) {
                var propName = prop.slice(3, 4).toLowerCase() + prop.slice(4, prop.length);
                Object.defineProperty(this, propName, {
                    get: this[prop],
                    enumerable: true,
                    configurable: true
                });
            }
            else if (prop.indexOf("set") == 0) {
                var propName = prop.slice(3, 4).toLowerCase() + prop.slice(4, prop.length);
                Object.defineProperty(this, propName, {
                    set: this[prop],
                    enumerable: true,
                    configurable: true
                });
            }
            else if (prop.indexOf("is") == 0) {
                var propName = prop.slice(2, 3).toLowerCase() + prop.slice(3, prop.length);
                Object.defineProperty(this, propName, {
                    get: this[prop],
                    enumerable: true,
                    configurable: true
                });
            }
        }
    }
});


var Tile = $Class.$extend({

    _x: 0,
    _y: 0,
    _value: 2,
    _lastMerge: -1,
    _element: null,
    _visible: false,

    __init__: function(x, y, value) {
        this.$super();
        this._element = document.createElement("div");
        e_board.appendChild(this._element);

        this._x = x;
        this._y = y;
        this._value = value || 2;
        this._update();

        setTimeout(function() {
            this.visible = true;
        }.bind(this), 100);

    },

    getX: function() {
        return this._x;
    },

    setX: function(x) {
        this._x = x;
        this._update();
    },

    getY: function() {
        return this._y;
    },

    setY: function(y) {
        this._y = y;
        this._update();
    },

    getValue: function() {
        return this._value;
    },

    setValue: function(value) {
        this._value = value;
        this._update();
    },

    isVisible: function() {
        return this._visible;
    },

    setVisible: function(visible) {
        this._visible = visible;
        this._update();
    },

    getLastMerge: function() {
        return this._lastMerge;
    },

    setLastMerge: function(iteration) {
        this._lastMerge = iteration;
    },

    destroy: function() {
        this.visible = false;
        setTimeout(function() {
            this._element.parentNode.removeChild(this._element);
        }.bind(this), 200);
    },

    _update: function() {
        var style = " tile-style-" + this.value;
        var pos = " tile-pos-" + this.x + "-" + this.y;
        var visible = (this._visible) ? " visible" : "";
        this._element.className = "tile" + style + pos + visible;
        while (this._element.hasChildNodes()) {
            this._element.removeChild(this._element.lastChild);
        }
        this._element.appendChild(document.createTextNode(this.value));
    }
});


var Game = $Class.$extend({

    GAME_STATUS_INGAME: "ingame",
    GAME_STATUS_LOOSE: "loose",
    GAME_STATUS_WIN: "win",

    gameStatus: null,

    _tiles: null,
    _iteration: 0,

    __init__: function() {
        this.$super();
        document.addEventListener("keydown", this._onKeyDown.bind(this), false);
        this.newGame();
    },

    _snapshots: null,

    getSnapshots: function() {
        return this._snapshots;
    },

    getGrid: function() {
        var grid = [];

        for (var y=0 ; y<4 ; y++) {
            grid.push([0, 0, 0, 0]);
        }

        for (var i=0 ; i<this._tiles.length ; i++) {
            grid[this._tiles[i].y][this._tiles[i].x] = this._tiles[i].value;
        }

        return grid;
    },

    getTile: function(x, y) {
        for (var i=0 ; i<this._tiles.length ; i++) {
            if (this._tiles[i].x == x && this._tiles[i].y == y) {
                return this._tiles[i];
            }
        }
        return null;
    },

    takeSnapshot: function() {
        this._snapshots.push(this.grid);
    },

    newGame: function() {
        this._iteration = 0;
        this._snapshots = [];
        e_board.className = "board";
        this.gameStatus = this.GAME_STATUS_INGAME;

        if (this._tiles) {
            for (var i=0 ; i<this._tiles.length ; i++) {
                this._tiles[i].destroy();
            }
        }
        this._tiles = [];

        this._appendTile();
        this._appendTile();

        this.takeSnapshot();
    },

    moveUp: function() {
        this._play(0, -1);
    },

    moveDown: function() {
        this._play(0, 1);
    },

    moveLeft: function() {
        this._play(-1, 0);
    },

    moveRight: function() {
        this._play(1, 0);
    },

    _play: function(dx, dy) {
        if (this.gameStatus != this.GAME_STATUS_INGAME) {
            return;
        }

        var moved = false;
        var grid = this.grid;
        var nx = 0;
        var ny = 0;
        var tile = null;

        for (var y=((dy<0)?0:3) ; ((dy<0)?y<=3:y>=0) ; y-=dy||1) {
            for (var x=((dx<0)?0:3) ; ((dx<0)?x<=3:x>=0) ; x-=dx||1) {

                if (grid[y][x] == 0) {
                    continue;
                }

                tile = this.getTile(x, y);

                for (var n=1 ; n<=3 ; n++) {
                    nx = Math.min(3, Math.max(0, x + n*dx));
                    ny = Math.min(3, Math.max(0, y + n*dy));

                    if (nx == tile.x && ny == tile.y) {
                        break;
                    }

                    if (grid[ny][nx] == 0) {
                        moved = true;
                        tile.x = nx;
                        tile.y = ny;
                        grid = this.grid;
                    }
                    else if (grid[ny][nx] == tile.value && this.getTile(nx, ny).lastMerge < this._iteration) {
                        moved = true;
                        this._removeTile(nx, ny);
                        tile.x = nx;
                        tile.y = ny;
                        tile.value *= 2;
                        tile.lastMerge = this._iteration;
                        grid = this.grid;
                        break;
                    }
                    else {
                        break;
                    }
                }

            }
        }
        this.takeSnapshot();

        if (moved) {
            this._appendTile();
            this.takeSnapshot();
        }

        this._updateGameStatus();

        switch (this.gameStatus) {
            case this.GAME_STATUS_LOOSE:
                e_board.className = "board gameover";
                break;
            case this.GAME_STATUS_WIN:
                e_board.className = "board goodgame";
                break;
        }

        this._iteration++;
    },

    _getEmptyCells: function() {
        var result = [];
        var grid = this.grid;

        for (var y=0 ; y<4 ; y++) {
            for (var x=0 ; x<4 ; x++) {
                if (grid[y][x] == 0) {
                    result.push({x: x, y: y});
                }
            }
        }

        return result;
    },

    _appendTile: function() {
        var value = Math.round(Math.random() * 2) * 2;
        var emptyCells = this._getEmptyCells();
        if (emptyCells.length) {
            var coord = emptyCells[Math.round(Math.random() * (emptyCells.length - 1))];
            var tile = new Tile(coord.x, coord.y, value);
            this._tiles.push(tile);
        }
    },

    _removeTile: function(x, y) {
        for (var i=0 ; i<this._tiles.length ; i++) {
            if (this._tiles[i].x == x && this._tiles[i].y == y) {
                this._tiles[i].destroy();
                this._tiles.splice(i, 1);
            }
        }
    },

    _updateGameStatus: function() {
        for (var i=0 ; i<this._tiles.length ; i++) {
            if (this._tiles[i].value == 2048) {
                this.gameStatus = this.GAME_STATUS_WIN;
                return;
            }
        }
        if (this._getEmptyCells().length == 0) {
            var grid = this.grid;
            for (var y=0 ; y<4 ; y++) {
                for (var x=0 ; x<4 ; x++) {
                    if (x < 3 && grid[y][x] == grid[y][x+1] || y < 3 && grid[y][x] == grid[y+1][x]) {
                        return;
                    }
                }
            }
            this.gameStatus = this.GAME_STATUS_LOOSE;
        }
    },

    _onKeyDown: function(event) {
        switch (event.keyCode) {
            case 38:
                this.moveUp();
                break;
            case 40:
                this.moveDown();
                break;
            case 37:
                this.moveLeft();
                break;
            case 39:
                this.moveRight();
                break;
        }
    }

});


var game = new Game();
